# Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import http.client
import uuid

from testtools import matchers

from keystone.common import provider_api
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit.ksfixtures import database
from keystone.tests.unit import test_v3

PROVIDERS = provider_api.ProviderAPIs


class CatalogTestCase(test_v3.RestfulTestCase):
    """Test service & endpoint CRUD."""

    # region crud tests

    def test_create_region_with_id(self):
        """Call ``PUT /regions/{region_id}`` w/o an ID in the request body."""
        ref = unit.new_region_ref()
        region_id = ref.pop('id')
        r = self.put(
            f'/regions/{region_id}',
            body={'region': ref},
            expected_status=http.client.CREATED,
        )
        self.assertValidRegionResponse(r, ref)
        # Double-check that the region ID was kept as-is and not
        # populated with a UUID, as is the case with POST /v3/regions
        self.assertEqual(region_id, r.json['region']['id'])

    def test_create_region_with_matching_ids(self):
        """Call ``PUT /regions/{region_id}`` with an ID in the request body."""
        ref = unit.new_region_ref()
        region_id = ref['id']
        r = self.put(
            f'/regions/{region_id}',
            body={'region': ref},
            expected_status=http.client.CREATED,
        )
        self.assertValidRegionResponse(r, ref)
        # Double-check that the region ID was kept as-is and not
        # populated with a UUID, as is the case with POST /v3/regions
        self.assertEqual(region_id, r.json['region']['id'])

    def test_create_region_with_duplicate_id(self):
        """Call ``PUT /regions/{region_id}``."""
        ref = unit.new_region_ref()
        region_id = ref['id']
        self.put(
            f'/regions/{region_id}',
            body={'region': ref},
            expected_status=http.client.CREATED,
        )
        # Create region again with duplicate id
        self.put(
            f'/regions/{region_id}',
            body={'region': ref},
            expected_status=http.client.CONFLICT,
        )

    def test_create_region(self):
        """Call ``POST /regions`` with an ID in the request body."""
        # the ref will have an ID defined on it
        ref = unit.new_region_ref()
        r = self.post('/regions', body={'region': ref})
        self.assertValidRegionResponse(r, ref)

        # we should be able to get the region, having defined the ID ourselves
        r = self.get('/regions/{region_id}'.format(region_id=ref['id']))
        self.assertValidRegionResponse(r, ref)

    def test_create_region_with_empty_id(self):
        """Call ``POST /regions`` with an empty ID in the request body."""
        ref = unit.new_region_ref(id='')

        r = self.post('/regions', body={'region': ref})
        self.assertValidRegionResponse(r, ref)
        self.assertNotEmpty(r.result['region'].get('id'))

    def test_create_region_without_id(self):
        """Call ``POST /regions`` without an ID in the request body."""
        ref = unit.new_region_ref()

        # instead of defining the ID ourselves...
        del ref['id']

        # let the service define the ID
        r = self.post('/regions', body={'region': ref})
        self.assertValidRegionResponse(r, ref)

    def test_create_region_without_description(self):
        """Call ``POST /regions`` without description in the request body."""
        ref = unit.new_region_ref(description=None)

        del ref['description']

        r = self.post('/regions', body={'region': ref})
        # Create the description in the reference to compare to since the
        # response should now have a description, even though we didn't send
        # it with the original reference.
        ref['description'] = ''
        self.assertValidRegionResponse(r, ref)

    def test_create_regions_with_same_description_string(self):
        """Call ``POST /regions`` with duplicate descriptions."""
        # NOTE(lbragstad): Make sure we can create two regions that have the
        # same description.
        region_desc = 'Some Region Description'

        ref1 = unit.new_region_ref(description=region_desc)
        ref2 = unit.new_region_ref(description=region_desc)

        resp1 = self.post('/regions', body={'region': ref1})
        self.assertValidRegionResponse(resp1, ref1)

        resp2 = self.post('/regions', body={'region': ref2})
        self.assertValidRegionResponse(resp2, ref2)

    def test_create_regions_without_descriptions(self):
        """Call ``POST /regions`` with no description."""
        # NOTE(lbragstad): Make sure we can create two regions that have
        # no description in the request body. The description should be
        # populated by Catalog Manager.
        ref1 = unit.new_region_ref()
        ref2 = unit.new_region_ref()

        del ref1['description']
        ref2['description'] = None

        resp1 = self.post('/regions', body={'region': ref1})

        resp2 = self.post('/regions', body={'region': ref2})
        # Create the descriptions in the references to compare to since the
        # responses should now have descriptions, even though we didn't send
        # a description with the original references.
        ref1['description'] = ''
        ref2['description'] = ''
        self.assertValidRegionResponse(resp1, ref1)
        self.assertValidRegionResponse(resp2, ref2)

    def test_create_region_with_conflicting_ids(self):
        """Call ``PUT /regions/{region_id}`` with conflicting region IDs."""
        # the region ref is created with an ID
        ref = unit.new_region_ref()

        # but instead of using that ID, make up a new, conflicting one
        self.put(
            f'/regions/{uuid.uuid4().hex}',
            body={'region': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_list_head_regions(self):
        """Call ``GET & HEAD /regions``."""
        resource_url = '/regions'
        r = self.get(resource_url)
        self.assertValidRegionListResponse(r, ref=self.region)
        self.head(resource_url, expected_status=http.client.OK)

    def _create_region_with_parent_id(self, parent_id=None):
        ref = unit.new_region_ref(parent_region_id=parent_id)
        return self.post('/regions', body={'region': ref})

    def test_list_regions_filtered_by_parent_region_id(self):
        """Call ``GET /regions?parent_region_id={parent_region_id}``."""
        new_region = self._create_region_with_parent_id()
        parent_id = new_region.result['region']['id']

        new_region = self._create_region_with_parent_id(parent_id)
        new_region = self._create_region_with_parent_id(parent_id)

        r = self.get(f'/regions?parent_region_id={parent_id}')

        for region in r.result['regions']:
            self.assertEqual(parent_id, region['parent_region_id'])

    def test_get_head_region(self):
        """Call ``GET & HEAD /regions/{region_id}``."""
        resource_url = f'/regions/{self.region_id}'
        r = self.get(resource_url)
        self.assertValidRegionResponse(r, self.region)
        self.head(resource_url, expected_status=http.client.OK)

    def test_update_region(self):
        """Call ``PATCH /regions/{region_id}``."""
        region = unit.new_region_ref()
        del region['id']
        r = self.patch(f'/regions/{self.region_id}', body={'region': region})
        self.assertValidRegionResponse(r, region)

    def test_update_region_without_description_keeps_original(self):
        """Call ``PATCH /regions/{region_id}``."""
        region_ref = unit.new_region_ref()

        resp = self.post('/regions', body={'region': region_ref})

        region_updates = {
            # update with something that's not the description
            'parent_region_id': self.region_id
        }
        resp = self.patch(
            '/regions/{}'.format(region_ref['id']),
            body={'region': region_updates},
        )

        # NOTE(dstanek): Keystone should keep the original description.
        self.assertEqual(
            region_ref['description'], resp.result['region']['description']
        )

    def test_update_region_with_null_description(self):
        """Call ``PATCH /regions/{region_id}``."""
        region = unit.new_region_ref(description=None)
        del region['id']
        r = self.patch(f'/regions/{self.region_id}', body={'region': region})

        # NOTE(dstanek): Keystone should turn the provided None value into
        # an empty string before storing in the backend.
        region['description'] = ''
        self.assertValidRegionResponse(r, region)

    def test_delete_region(self):
        """Call ``DELETE /regions/{region_id}``."""
        ref = unit.new_region_ref()
        r = self.post('/regions', body={'region': ref})
        self.assertValidRegionResponse(r, ref)

        self.delete('/regions/{region_id}'.format(region_id=ref['id']))

    # service crud tests

    def test_create_service(self):
        """Call ``POST /services``."""
        ref = unit.new_service_ref()
        r = self.post('/services', body={'service': ref})
        self.assertValidServiceResponse(r, ref)

    def test_create_service_no_name(self):
        """Call ``POST /services``."""
        ref = unit.new_service_ref()
        del ref['name']
        r = self.post('/services', body={'service': ref})
        ref['name'] = ''
        self.assertValidServiceResponse(r, ref)

    def test_create_service_no_enabled(self):
        """Call ``POST /services``."""
        ref = unit.new_service_ref()
        del ref['enabled']
        r = self.post('/services', body={'service': ref})
        ref['enabled'] = True
        self.assertValidServiceResponse(r, ref)
        self.assertIs(True, r.result['service']['enabled'])

    def test_create_service_enabled_false(self):
        """Call ``POST /services``."""
        ref = unit.new_service_ref(enabled=False)
        r = self.post('/services', body={'service': ref})
        self.assertValidServiceResponse(r, ref)
        self.assertIs(False, r.result['service']['enabled'])

    def test_create_service_enabled_true(self):
        """Call ``POST /services``."""
        ref = unit.new_service_ref(enabled=True)
        r = self.post('/services', body={'service': ref})
        self.assertValidServiceResponse(r, ref)
        self.assertIs(True, r.result['service']['enabled'])

    def test_create_service_enabled_str_random(self):
        """Call ``POST /services``."""
        ref = unit.new_service_ref(enabled='puppies')
        self.post(
            '/services',
            body={'service': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_list_head_services(self):
        """Call ``GET & HEAD /services``."""
        resource_url = '/services'
        r = self.get(resource_url)
        self.assertValidServiceListResponse(r, ref=self.service)
        self.head(resource_url, expected_status=http.client.OK)

    def _create_random_service(self):
        ref = unit.new_service_ref()
        response = self.post('/services', body={'service': ref})
        return response.json['service']

    def test_filter_list_services_by_type(self):
        """Call ``GET /services?type=<some type>``."""
        target_ref = self._create_random_service()

        # create unrelated services
        self._create_random_service()
        self._create_random_service()

        response = self.get('/services?type=' + target_ref['type'])
        self.assertValidServiceListResponse(response, ref=target_ref)

        filtered_service_list = response.json['services']
        self.assertEqual(1, len(filtered_service_list))

        filtered_service = filtered_service_list[0]
        self.assertEqual(target_ref['type'], filtered_service['type'])

    def test_filter_list_services_by_name(self):
        """Call ``GET /services?name=<some name>``."""
        # create unrelated services
        self._create_random_service()
        self._create_random_service()

        # create the desired service
        target_ref = self._create_random_service()

        response = self.get('/services?name=' + target_ref['name'])
        self.assertValidServiceListResponse(response, ref=target_ref)

        filtered_service_list = response.json['services']
        self.assertEqual(1, len(filtered_service_list))

        filtered_service = filtered_service_list[0]
        self.assertEqual(target_ref['name'], filtered_service['name'])

    def test_filter_list_services_by_name_with_list_limit(self):
        """Call ``GET /services?name=<some name>``."""
        self.config_fixture.config(list_limit=1)

        self.test_filter_list_services_by_name()

    def test_get_head_service(self):
        """Call ``GET & HEAD /services/{service_id}``."""
        resource_url = f'/services/{self.service_id}'
        r = self.get(resource_url)
        self.assertValidServiceResponse(r, self.service)
        self.head(resource_url, expected_status=http.client.OK)

    def test_update_service(self):
        """Call ``PATCH /services/{service_id}``."""
        service = unit.new_service_ref()
        del service['id']
        r = self.patch(
            f'/services/{self.service_id}', body={'service': service}
        )
        self.assertValidServiceResponse(r, service)

    def test_delete_service(self):
        """Call ``DELETE /services/{service_id}``."""
        self.delete(f'/services/{self.service_id}')

    # endpoint crud tests

    def test_list_head_endpoints(self):
        """Call ``GET & HEAD /endpoints``."""
        resource_url = '/endpoints'
        r = self.get(resource_url)
        self.assertValidEndpointListResponse(r, ref=self.endpoint)
        self.head(resource_url, expected_status=http.client.OK)

    def _create_random_endpoint(
        self, interface='public', parent_region_id=None
    ):
        region = self._create_region_with_parent_id(parent_id=parent_region_id)
        service = self._create_random_service()
        ref = unit.new_endpoint_ref(
            service_id=service['id'],
            interface=interface,
            region_id=region.result['region']['id'],
        )

        response = self.post('/endpoints', body={'endpoint': ref})
        return response.json['endpoint']

    def test_list_endpoints_filtered_by_interface(self):
        """Call ``GET /endpoints?interface={interface}``."""
        ref = self._create_random_endpoint(interface='internal')

        response = self.get('/endpoints?interface={}'.format(ref['interface']))
        self.assertValidEndpointListResponse(response, ref=ref)

        for endpoint in response.json['endpoints']:
            self.assertEqual(ref['interface'], endpoint['interface'])

    def test_list_endpoints_filtered_by_service_id(self):
        """Call ``GET /endpoints?service_id={service_id}``."""
        ref = self._create_random_endpoint()

        response = self.get(
            '/endpoints?service_id={}'.format(ref['service_id'])
        )
        self.assertValidEndpointListResponse(response, ref=ref)

        for endpoint in response.json['endpoints']:
            self.assertEqual(ref['service_id'], endpoint['service_id'])

    def test_list_endpoints_filtered_by_region_id(self):
        """Call ``GET /endpoints?region_id={region_id}``."""
        ref = self._create_random_endpoint()

        response = self.get('/endpoints?region_id={}'.format(ref['region_id']))
        self.assertValidEndpointListResponse(response, ref=ref)

        for endpoint in response.json['endpoints']:
            self.assertEqual(ref['region_id'], endpoint['region_id'])

    def test_list_endpoints_filtered_by_parent_region_id(self):
        """Call ``GET /endpoints?region_id={region_id}``.

        Ensure passing the parent_region_id as filter returns an
        empty list.

        """
        parent_region = self._create_region_with_parent_id()
        parent_region_id = parent_region.result['region']['id']
        self._create_random_endpoint(parent_region_id=parent_region_id)

        response = self.get(f'/endpoints?region_id={parent_region_id}')
        self.assertEqual(0, len(response.json['endpoints']))

    def test_list_endpoints_with_multiple_filters(self):
        """Call ``GET /endpoints?interface={interface}...``.

        Ensure passing different combinations of interface, region_id and
        service_id as filters will return the correct result.

        """
        # interface and region_id specified
        ref = self._create_random_endpoint(interface='internal')
        response = self.get(
            '/endpoints?interface={}&region_id={}'.format(
                ref['interface'], ref['region_id']
            )
        )
        self.assertValidEndpointListResponse(response, ref=ref)

        for endpoint in response.json['endpoints']:
            self.assertEqual(ref['interface'], endpoint['interface'])
            self.assertEqual(ref['region_id'], endpoint['region_id'])

        # interface and service_id specified
        ref = self._create_random_endpoint(interface='internal')
        response = self.get(
            '/endpoints?interface={}&service_id={}'.format(
                ref['interface'], ref['service_id']
            )
        )
        self.assertValidEndpointListResponse(response, ref=ref)

        for endpoint in response.json['endpoints']:
            self.assertEqual(ref['interface'], endpoint['interface'])
            self.assertEqual(ref['service_id'], endpoint['service_id'])

        # region_id and service_id specified
        ref = self._create_random_endpoint(interface='internal')
        response = self.get(
            '/endpoints?region_id={}&service_id={}'.format(
                ref['region_id'], ref['service_id']
            )
        )
        self.assertValidEndpointListResponse(response, ref=ref)

        for endpoint in response.json['endpoints']:
            self.assertEqual(ref['region_id'], endpoint['region_id'])
            self.assertEqual(ref['service_id'], endpoint['service_id'])

        # interface, region_id and service_id specified
        ref = self._create_random_endpoint(interface='internal')
        response = self.get(
            ('/endpoints?interface={}&region_id={}&service_id={}').format(
                ref['interface'], ref['region_id'], ref['service_id']
            )
        )
        self.assertValidEndpointListResponse(response, ref=ref)

        for endpoint in response.json['endpoints']:
            self.assertEqual(ref['interface'], endpoint['interface'])
            self.assertEqual(ref['region_id'], endpoint['region_id'])
            self.assertEqual(ref['service_id'], endpoint['service_id'])

    def test_list_endpoints_with_random_filter_values(self):
        """Call ``GET /endpoints?interface={interface}...``.

        Ensure passing random values for: interface, region_id and
        service_id will return an empty list.

        """
        self._create_random_endpoint(interface='internal')

        response = self.get(f'/endpoints?interface=admin')
        self.assertEqual(0, len(response.json['endpoints']))

        response = self.get(f'/endpoints?region_id={uuid.uuid4().hex}')
        self.assertEqual(0, len(response.json['endpoints']))

        response = self.get(f'/endpoints?service_id={uuid.uuid4().hex}')
        self.assertEqual(0, len(response.json['endpoints']))

    def test_create_endpoint_no_enabled(self):
        """Call ``POST /endpoints``."""
        ref = unit.new_endpoint_ref(
            service_id=self.service_id,
            interface='public',
            region_id=self.region_id,
        )
        r = self.post('/endpoints', body={'endpoint': ref})
        ref['enabled'] = True
        self.assertValidEndpointResponse(r, ref)

    def test_create_endpoint_enabled_true(self):
        """Call ``POST /endpoints`` with enabled: true."""
        ref = unit.new_endpoint_ref(
            service_id=self.service_id,
            interface='public',
            region_id=self.region_id,
            enabled=True,
        )
        r = self.post('/endpoints', body={'endpoint': ref})
        self.assertValidEndpointResponse(r, ref)

    def test_create_endpoint_enabled_false(self):
        """Call ``POST /endpoints`` with enabled: false."""
        ref = unit.new_endpoint_ref(
            service_id=self.service_id,
            interface='public',
            region_id=self.region_id,
            enabled=False,
        )
        r = self.post('/endpoints', body={'endpoint': ref})
        self.assertValidEndpointResponse(r, ref)

    def test_create_endpoint_enabled_str_true(self):
        """Call ``POST /endpoints`` with enabled: 'True'."""
        ref = unit.new_endpoint_ref(
            service_id=self.service_id,
            interface='public',
            region_id=self.region_id,
            enabled='True',
        )
        self.post(
            '/endpoints',
            body={'endpoint': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_create_endpoint_enabled_str_false(self):
        """Call ``POST /endpoints`` with enabled: 'False'."""
        ref = unit.new_endpoint_ref(
            service_id=self.service_id,
            interface='public',
            region_id=self.region_id,
            enabled='False',
        )
        self.post(
            '/endpoints',
            body={'endpoint': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_create_endpoint_enabled_str_random(self):
        """Call ``POST /endpoints`` with enabled: 'puppies'."""
        ref = unit.new_endpoint_ref(
            service_id=self.service_id,
            interface='public',
            region_id=self.region_id,
            enabled='puppies',
        )
        self.post(
            '/endpoints',
            body={'endpoint': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_create_endpoint_with_invalid_region_id(self):
        """Call ``POST /endpoints``."""
        ref = unit.new_endpoint_ref(service_id=self.service_id)
        self.post(
            '/endpoints',
            body={'endpoint': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_create_endpoint_with_region(self):
        """EndpointV3 creates the region before creating the endpoint.

        This occurs when endpoint is provided with 'region' and no 'region_id'.
        """
        ref = unit.new_endpoint_ref_with_region(
            service_id=self.service_id, region=uuid.uuid4().hex
        )
        self.post('/endpoints', body={'endpoint': ref})
        # Make sure the region is created
        self.get('/regions/{region_id}'.format(region_id=ref["region"]))

    def test_create_endpoint_with_no_region(self):
        """EndpointV3 allows to creates the endpoint without region."""
        ref = unit.new_endpoint_ref(service_id=self.service_id, region_id=None)
        del ref['region_id']  # cannot just be None, it needs to not exist
        self.post('/endpoints', body={'endpoint': ref})

    def test_create_endpoint_with_empty_url(self):
        """Call ``POST /endpoints``."""
        ref = unit.new_endpoint_ref(service_id=self.service_id, url='')
        self.post(
            '/endpoints',
            body={'endpoint': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_get_head_endpoint(self):
        """Call ``GET & HEAD /endpoints/{endpoint_id}``."""
        resource_url = f'/endpoints/{self.endpoint_id}'
        r = self.get(resource_url)
        self.assertValidEndpointResponse(r, self.endpoint)
        self.head(resource_url, expected_status=http.client.OK)

    def test_update_endpoint(self):
        """Call ``PATCH /endpoints/{endpoint_id}``."""
        ref = unit.new_endpoint_ref(
            service_id=self.service_id,
            interface='public',
            region_id=self.region_id,
        )
        del ref['id']
        r = self.patch(
            f'/endpoints/{self.endpoint_id}', body={'endpoint': ref}
        )
        ref['enabled'] = True
        self.assertValidEndpointResponse(r, ref)

    def test_update_endpoint_enabled_true(self):
        """Call ``PATCH /endpoints/{endpoint_id}`` with enabled: True."""
        r = self.patch(
            f'/endpoints/{self.endpoint_id}',
            body={'endpoint': {'enabled': True}},
        )
        self.assertValidEndpointResponse(r, self.endpoint)

    def test_update_endpoint_enabled_false(self):
        """Call ``PATCH /endpoints/{endpoint_id}`` with enabled: False."""
        r = self.patch(
            f'/endpoints/{self.endpoint_id}',
            body={'endpoint': {'enabled': False}},
        )
        exp_endpoint = copy.copy(self.endpoint)
        exp_endpoint['enabled'] = False
        self.assertValidEndpointResponse(r, exp_endpoint)

    def test_update_endpoint_enabled_str_true(self):
        """Call ``PATCH /endpoints/{endpoint_id}`` with enabled: 'True'."""
        self.patch(
            f'/endpoints/{self.endpoint_id}',
            body={'endpoint': {'enabled': 'True'}},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_endpoint_enabled_str_false(self):
        """Call ``PATCH /endpoints/{endpoint_id}`` with enabled: 'False'."""
        self.patch(
            f'/endpoints/{self.endpoint_id}',
            body={'endpoint': {'enabled': 'False'}},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_endpoint_enabled_str_random(self):
        """Call ``PATCH /endpoints/{endpoint_id}`` with enabled: 'kitties'."""
        self.patch(
            f'/endpoints/{self.endpoint_id}',
            body={'endpoint': {'enabled': 'kitties'}},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_delete_endpoint(self):
        """Call ``DELETE /endpoints/{endpoint_id}``."""
        self.delete(f'/endpoints/{self.endpoint_id}')

    def test_deleting_endpoint_with_space_in_url(self):
        # add a space to all urls (intentional "i d" to test bug)
        url_with_space = "http://127.0.0.1:8774 /v1.1/\\$(tenant_i d)s"

        # create a v3 endpoint ref
        ref = unit.new_endpoint_ref(
            service_id=self.service['id'],
            region_id=None,
            publicurl=url_with_space,
            internalurl=url_with_space,
            adminurl=url_with_space,
            url=url_with_space,
        )

        # add the endpoint to the database
        PROVIDERS.catalog_api.create_endpoint(ref['id'], ref)

        # delete the endpoint
        self.delete('/endpoints/{}'.format(ref['id']))

        # make sure it's deleted (GET should return Not Found)
        self.get(
            '/endpoints/{}'.format(ref['id']),
            expected_status=http.client.NOT_FOUND,
        )

    def test_endpoint_create_with_valid_url(self):
        """Create endpoint with valid url should be tested,too."""
        # list one valid url is enough, no need to list too much
        valid_url = 'http://127.0.0.1:8774/v1.1/$(project_id)s'

        ref = unit.new_endpoint_ref(
            self.service_id,
            interface='public',
            region_id=self.region_id,
            url=valid_url,
        )
        self.post('/endpoints', body={'endpoint': ref})

    def test_endpoint_create_with_valid_url_project_id(self):
        """Create endpoint with valid url should be tested,too."""
        valid_url = 'http://127.0.0.1:8774/v1.1/$(project_id)s'

        ref = unit.new_endpoint_ref(
            self.service_id,
            interface='public',
            region_id=self.region_id,
            url=valid_url,
        )
        self.post('/endpoints', body={'endpoint': ref})

    def test_endpoint_create_with_invalid_url(self):
        """Test the invalid cases: substitutions is not exactly right."""
        invalid_urls = [
            # using a substitution that is not whitelisted - KeyError
            'http://127.0.0.1:8774/v1.1/$(nonexistent)s',
            # invalid formatting - ValueError
            'http://127.0.0.1:8774/v1.1/$(project_id)',
            'http://127.0.0.1:8774/v1.1/$(project_id)t',
            'http://127.0.0.1:8774/v1.1/$(project_id',
            # invalid type specifier - TypeError
            # admin_url is a string not an int
            'http://127.0.0.1:8774/v1.1/$(admin_url)d',
        ]

        ref = unit.new_endpoint_ref(self.service_id)

        for invalid_url in invalid_urls:
            ref['url'] = invalid_url
            self.post(
                '/endpoints',
                body={'endpoint': ref},
                expected_status=http.client.BAD_REQUEST,
            )


class TestMultiRegion(test_v3.RestfulTestCase):
    def test_catalog_with_multi_region_reports_all_endpoints(self):
        # Create two separate regions
        first_region = self.post(
            '/regions', body={'region': unit.new_region_ref()}
        ).json_body['region']

        second_region = self.post(
            '/regions', body={'region': unit.new_region_ref()}
        ).json_body['region']

        # Create two services with the same type but separate name.
        first_service = self.post(
            '/services', body={'service': unit.new_service_ref(type='foobar')}
        ).json_body['service']

        second_service = self.post(
            '/services', body={'service': unit.new_service_ref(type='foobar')}
        ).json_body['service']

        # Create an endpoint for each service
        first_endpoint = self.post(
            '/endpoints',
            body={
                'endpoint': unit.new_endpoint_ref(
                    first_service['id'],
                    interface='public',
                    region_id=first_region['id'],
                )
            },
        ).json_body['endpoint']

        second_endpoint = self.post(
            '/endpoints',
            body={
                'endpoint': unit.new_endpoint_ref(
                    second_service['id'],
                    interface='public',
                    region_id=second_region['id'],
                )
            },
        ).json_body['endpoint']

        # Assert the endpoints and services from each region are in the
        # catalog.
        found_first_endpoint = False
        found_second_endpoint = False
        catalog = self.get('/auth/catalog/').json_body['catalog']
        for service in catalog:
            if service['id'] == first_service['id']:
                endpoint = service['endpoints'][0]
                self.assertEqual(endpoint['id'], first_endpoint['id'])
                self.assertEqual(endpoint['region_id'], first_region['id'])
                found_first_endpoint = True
            elif service['id'] == second_service['id']:
                endpoint = service['endpoints'][0]
                self.assertEqual(endpoint['id'], second_endpoint['id'])
                self.assertEqual(endpoint['region_id'], second_region['id'])
                found_second_endpoint = True

        self.assertTrue(found_first_endpoint)
        self.assertTrue(found_second_endpoint)


class TestCatalogAPISQL(unit.TestCase):
    """Test for the catalog Manager against the SQL backend."""

    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.load_backends()

        service = unit.new_service_ref()
        self.service_id = service['id']
        PROVIDERS.catalog_api.create_service(self.service_id, service)

        self.create_endpoint(service_id=self.service_id)

        PROVIDERS.resource_api.create_domain(
            default_fixtures.ROOT_DOMAIN['id'], default_fixtures.ROOT_DOMAIN
        )

    def create_endpoint(self, service_id, **kwargs):
        endpoint = unit.new_endpoint_ref(
            service_id=service_id, region_id=None, **kwargs
        )

        PROVIDERS.catalog_api.create_endpoint(endpoint['id'], endpoint)
        return endpoint

    def config_overrides(self):
        super().config_overrides()
        self.config_fixture.config(group='catalog', driver='sql')

    def test_get_catalog_ignores_endpoints_with_invalid_urls(self):
        user_id = uuid.uuid4().hex

        # create a project since the project should exist if we want to
        # filter the catalog by the project or replace the url with a
        # valid project id.
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        project = unit.new_project_ref(domain_id=domain['id'])
        PROVIDERS.resource_api.create_project(project['id'], project)

        # the only endpoint in the catalog is the one created in setUp
        catalog = PROVIDERS.catalog_api.get_v3_catalog(user_id, project['id'])
        self.assertEqual(1, len(catalog[0]['endpoints']))
        # it's also the only endpoint in the backend
        self.assertEqual(1, len(PROVIDERS.catalog_api.list_endpoints()))

        # create a new, invalid endpoint - malformed type declaration
        self.create_endpoint(
            self.service_id, url='http://keystone/%(project_id)'
        )

        # create a new, invalid endpoint - nonexistent key
        self.create_endpoint(
            self.service_id, url='http://keystone/%(you_wont_find_me)s'
        )

        # verify that the invalid endpoints don't appear in the catalog
        catalog = PROVIDERS.catalog_api.get_v3_catalog(user_id, project['id'])
        self.assertEqual(1, len(catalog[0]['endpoints']))
        # all three appear in the backend
        self.assertEqual(3, len(PROVIDERS.catalog_api.list_endpoints()))

        # create another valid endpoint - project_id will be replaced
        self.create_endpoint(
            self.service_id, url='http://keystone/%(project_id)s'
        )

        # there are two valid endpoints, positive check
        catalog = PROVIDERS.catalog_api.get_v3_catalog(user_id, project['id'])
        self.assertThat(catalog[0]['endpoints'], matchers.HasLength(2))

        # If the URL has no 'project_id' to substitute, we will skip the
        # endpoint which contains this kind of URL, negative check.
        project_id = None
        catalog = PROVIDERS.catalog_api.get_v3_catalog(user_id, project_id)
        self.assertThat(catalog[0]['endpoints'], matchers.HasLength(1))

    def test_get_catalog_always_returns_service_name(self):
        user_id = uuid.uuid4().hex
        # create a project since the project should exist if we want to
        # filter the catalog by the project or replace the url with a
        # valid project id.
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        project = unit.new_project_ref(domain_id=domain['id'])
        PROVIDERS.resource_api.create_project(project['id'], project)

        # create a service, with a name
        named_svc = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(named_svc['id'], named_svc)
        self.create_endpoint(service_id=named_svc['id'])

        # create a service, with no name
        unnamed_svc = unit.new_service_ref(name=None)
        del unnamed_svc['name']
        PROVIDERS.catalog_api.create_service(unnamed_svc['id'], unnamed_svc)
        self.create_endpoint(service_id=unnamed_svc['id'])

        catalog = PROVIDERS.catalog_api.get_v3_catalog(user_id, project['id'])

        named_endpoint = [
            ep for ep in catalog if ep['type'] == named_svc['type']
        ][0]
        self.assertEqual(named_svc['name'], named_endpoint['name'])

        unnamed_endpoint = [
            ep for ep in catalog if ep['type'] == unnamed_svc['type']
        ][0]
        self.assertEqual('', unnamed_endpoint['name'])


# TODO(dstanek): this needs refactoring with the test above, but we are in a
# crunch so that will happen in a future patch.
class TestCatalogAPISQLRegions(unit.TestCase):
    """Test for the catalog Manager against the SQL backend."""

    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.load_backends()
        PROVIDERS.resource_api.create_domain(
            default_fixtures.ROOT_DOMAIN['id'], default_fixtures.ROOT_DOMAIN
        )

    def config_overrides(self):
        super().config_overrides()
        self.config_fixture.config(group='catalog', driver='sql')

    def test_get_catalog_returns_proper_endpoints_with_no_region(self):
        service = unit.new_service_ref()
        service_id = service['id']
        PROVIDERS.catalog_api.create_service(service_id, service)

        endpoint = unit.new_endpoint_ref(service_id=service_id, region_id=None)
        del endpoint['region_id']
        PROVIDERS.catalog_api.create_endpoint(endpoint['id'], endpoint)

        # create a project since the project should exist if we want to
        # filter the catalog by the project or replace the url with a
        # valid project id.
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        project = unit.new_project_ref(domain_id=domain['id'])
        PROVIDERS.resource_api.create_project(project['id'], project)
        user_id = uuid.uuid4().hex
        catalog = PROVIDERS.catalog_api.get_v3_catalog(user_id, project['id'])
        self.assertValidCatalogEndpoint(
            catalog[0]['endpoints'][0], ref=endpoint
        )

    def test_get_catalog_returns_proper_endpoints_with_region(self):
        service = unit.new_service_ref()
        service_id = service['id']
        PROVIDERS.catalog_api.create_service(service_id, service)

        endpoint = unit.new_endpoint_ref(service_id=service_id)
        region = unit.new_region_ref(id=endpoint['region_id'])
        PROVIDERS.catalog_api.create_region(region)
        PROVIDERS.catalog_api.create_endpoint(endpoint['id'], endpoint)

        endpoint = PROVIDERS.catalog_api.get_endpoint(endpoint['id'])
        user_id = uuid.uuid4().hex
        # create a project since the project should exist if we want to
        # filter the catalog by the project or replace the url with a
        # valid project id.
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        project = unit.new_project_ref(domain_id=domain['id'])
        PROVIDERS.resource_api.create_project(project['id'], project)

        catalog = PROVIDERS.catalog_api.get_v3_catalog(user_id, project['id'])
        self.assertValidCatalogEndpoint(
            catalog[0]['endpoints'][0], ref=endpoint
        )

    def assertValidCatalogEndpoint(self, entity, ref=None):
        keys = ['description', 'id', 'interface', 'name', 'region_id', 'url']
        for k in keys:
            self.assertEqual(ref.get(k), entity[k], k)
        self.assertEqual(entity['region_id'], entity['region'])



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from keystone import exception


def federated_objects_to_list(fed_ref):
    """Create a new reformatted federated object list using the one passed in.

    When returning federated objects with a user we only need the attributes
    idp_id, protocol_id, and unique_id. Therefore, we pull these elements out
    of the fed_ref and create a newly formatted list with the needed
    information. We simply group each federated object's protocol_ids and
    unique_ids under the corresponding idp_id.

    :returns list: Containing the user's federated objects
    """
    if not fed_ref:
        return []

    fed = {}
    for fed_dict in fed_ref:
        fed.setdefault(
            fed_dict['idp_id'], {'idp_id': fed_dict['idp_id'], 'protocols': []}
        )['protocols'].append(
            {
                'protocol_id': fed_dict['protocol_id'],
                'unique_id': fed_dict['unique_id'],
            }
        )

    return list(fed.values())


class ShadowUsersDriverBase(metaclass=abc.ABCMeta):
    """Interface description for an Shadow Users driver."""

    @abc.abstractmethod
    def create_federated_object(self, fed_dict):
        """Create a new federated object.

        :param dict federated_dict: Reference to the federated user
        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def create_federated_user(self, domain_id, federated_dict, email=None):
        """Create a new user with the federated identity.

        :param domain_id: The domain ID of the IdP used for the federated user
        :param dict federated_dict: Reference to the federated user
        :param email: Federated user's email
        :returns dict: Containing the user reference

        """
        raise exception.NotImplemented()

    def delete_federated_object(self, user_id):
        """Delete a user's federated objects.

        :param user_id: Unique identifier of the user
        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def get_federated_objects(self, user_id):
        """Get all federated objects for a user.

        :param user_id: Unique identifier of the user
        :returns list: Containing the user's federated objects

        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def get_federated_user(self, idp_id, protocol_id, unique_id):
        """Return the found user for the federated identity.

        :param idp_id: The identity provider ID
        :param protocol_id: The federation protocol ID
        :param unique_id: The unique ID for the user
        :returns dict: Containing the user reference

        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def update_federated_user_display_name(
        self, idp_id, protocol_id, unique_id, display_name
    ):
        """Update federated user's display name if changed.

        :param idp_id: The identity provider ID
        :param protocol_id: The federation protocol ID
        :param unique_id: The unique ID for the user
        :param display_name: The user's display name

        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def get_user(self, user_id):
        """Return the found user.

        :param user_id: Unique identifier of the user
        :returns dict: Containing the user reference

        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def create_nonlocal_user(self, user_dict):
        """Create a new non-local user.

        :param dict user_dict: Reference to the non-local user
        :returns dict: Containing the user reference

        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def set_last_active_at(self, user_id):
        """Set the last active at date for the user.

        :param user_id: Unique identifier of the user

        """
        raise exception.NotImplemented()

    @abc.abstractmethod
    def list_federated_users_info(self, hints=None):
        """Get the shadow users info with the specified filters.

        :param hints: contains the list of filters yet to be satisfied.
                      Any filters satisfied here will be removed so that
                      the caller will know if any filters remain.
        :returns list: A list of objects that containing the shadow users
                       reference.

        """
        raise exception.NotImplemented()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# NOTE(morganfainberg) This file shouldn't have flake8 run on it as it has
# code examples that will fail normal CI pep8/flake8 tests. This is expected.
# The code has been moved here to ensure that proper tests occur on the
# test_hacking_checks test cases.
# flake8: noqa

import sys

import fixtures


class HackingCode(fixtures.Fixture):
    """A fixture to house the various code examples for the keystone hacking
    style checks.
    """

    mutable_default_args = {
        'code': """
                def f():
                    pass

                def f(a, b='', c=None):
                    pass

                def f(bad=[]):
                    pass

                def f(foo, bad=[], more_bad=[x for x in range(3)]):
                    pass

                def f(foo, bad={}):
                    pass

                def f(foo, bad={}, another_bad=[], fine=None):
                    pass

                def f(bad=[]): # noqa
                    pass

                def funcs(bad=dict(), more_bad=list(), even_more_bad=set()):
                    "creating mutables through builtins"

                def funcs(bad=something(), more_bad=some_object.something()):
                    "defaults from any functions"

                def f(bad=set(), more_bad={x for x in range(3)},
                       even_more_bad={1, 2, 3}):
                    "set and set comprehession"

                def f(bad={x: x for x in range(3)}):
                    "dict comprehension"
            """,
        'expected_errors': [
            (7, 10, 'K001'),
            (10, 15, 'K001'),
            (10, 28, 'K001'),
            (13, 15, 'K001'),
            (16, 15, 'K001'),
            (16, 31, 'K001'),
            (22, 14, 'K001'),
            (22, 31, 'K001'),
            (22, 53, 'K001'),
            (25, 14, 'K001'),
            (25, 36, 'K001'),
            (28, 10, 'K001'),
            (28, 26, 'K001'),
            (29, 21, 'K001'),
            (32, 10, 'K001'),
        ],
    }

    # NOTE(browne): This is gross, but in Python 3.4 and earlier, the ast
    # module returns the incorrect col_offset for two of the defined functions
    # in the code sample above.
    if sys.version_info < (3, 5):
        mutable_default_args['expected_errors'][12] = (28, 27, 'K001')
        mutable_default_args['expected_errors'][14] = (32, 11, 'K001')

    # NOTE(gmann): Python version < 3.8.2 (before Ubuntu Focal) returns
    # the incorrect col_offset for below function defined in above code sample
    # def f(foo, bad=[], more_bad=[x for x in range(3)]):
    if sys.version_info < (3, 8, 2):
        mutable_default_args['expected_errors'][2] = (10, 29, 'K001')  # type: ignore[index]

    comments_begin_with_space = {
        'code': """
            # This is a good comment

            #This is a bad one

            # This is alright and can
            #    be continued with extra indentation
            #    if that's what the developer wants.
        """,
        'expected_errors': [(3, 0, 'K002')],
    }

    asserting_none_equality = {
        'code': """
            class Test(object):

                def test(self):
                    self.assertEqual('', '')
                    self.assertEqual('', None)
                    self.assertEqual(None, '')
                    self.assertNotEqual('', None)
                    self.assertNotEqual(None, '')
                    self.assertNotEqual('', None) # noqa
                    self.assertNotEqual(None, '') # noqa
        """,
        'expected_errors': [
            (5, 8, 'K003'),
            (6, 8, 'K003'),
            (7, 8, 'K004'),
            (8, 8, 'K004'),
        ],
    }

    dict_constructor = {
        'code': """
            lower_res = {k.lower(): v for k, v in res[1].items()}
            fool = dict(a='a', b='b')
            lower_res = dict((k.lower(), v) for k, v in res[1].items())
            attrs = dict([(k, _from_json(v))])
            dict([[i,i] for i in range(3)])
            dict(({1:2}))
        """,
        'expected_errors': [(3, 0, 'K008'), (4, 0, 'K008'), (5, 0, 'K008')],
    }


class HackingTranslations(fixtures.Fixture):
    """Fixtures for checking translation rules.

    1. Exception messages should be translated
    2. Logging messages should not be translated
    3. If a message is used for both an exception and logging it
       should be translated
    """

    shared_imports = """
                import logging
                import logging as stlib_logging
                from keystone.i18n import _
                from keystone.i18n import _ as oslo_i18n
                from oslo_log import log
                from oslo_log import log as oslo_logging
    """

    examples = [
        {
            'code': """
                # stdlib logging
                LOG = logging.getLogger()
                LOG.info(_('text'))
                class C:
                    def __init__(self):
                        LOG.warning(oslo_i18n('text', {}))
            """,
            'expected_errors': [(3, 9, 'K005'), (6, 20, 'K005')],
        },
        {
            'code': """
                # stdlib logging w/ alias and specifying a logger
                class C:
                    def __init__(self):
                        self.L = logging.getLogger(__name__)
                    def m(self):
                        self.L.warning(
                            _('text'), {}
                        )
            """,
            'expected_errors': [(7, 12, 'K005')],
        },
        {
            'code': """
                # oslo logging and specifying a logger
                L = log.getLogger(__name__)
                L.error(oslo_i18n('text'))
            """,
            'expected_errors': [(3, 8, 'K005')],
        },
        {
            'code': """
                # oslo logging w/ alias
                class C:
                    def __init__(self):
                        self.LOG = oslo_logging.getLogger()
                        self.LOG.critical(_('text'))
            """,
            'expected_errors': [(5, 26, 'K005')],
        },
        {
            'code': """
                LOG = log.getLogger(__name__)
                # translation on a separate line
                msg = _('text')
                LOG.exception(msg)
            """,
            'expected_errors': [(4, 14, 'K005')],
        },
        {
            'code': """
                # this should be an error even if it'll be raised later.
                L = log.getLogger(__name__)
                msg = _('text')
                L.warning(msg)
                raise Exception(msg)
            """,
            'expected_errors': [(4, 10, 'K005')],
        },
        {
            'code': """
                L = log.getLogger(__name__)
                def f():
                    msg = _('text')
                    L.warning(msg)
                    something = True  # add an extra statement here
                    raise Exception(msg)
            """,
            'expected_errors': [(4, 14, 'K005')],
        },
        {
            'code': """
                LOG = log.getLogger(__name__)
                def func():
                    msg = _('text')
                    LOG.warning(msg)
                    raise Exception('some other message')
            """,
            'expected_errors': [(4, 16, 'K005')],
        },
        {
            'code': """
                LOG = log.getLogger(__name__)
                if True:
                    msg = _('text')
                else:
                    msg = _('text')
                LOG.warning(msg)
                raise Exception(msg)
            """,
            'expected_errors': [(6, 12, 'K005')],
        },
        {
            'code': """
                LOG = log.getLogger(__name__)
                if True:
                    msg = _('text')
                else:
                    msg = _('text')
                LOG.warning(msg)
            """,
            'expected_errors': [(6, 12, 'K005')],
        },
        {
            'code': """
                LOG = log.getLogger(__name__)
                msg = _LW('text')
                LOG.warning(msg)
                msg = _('something else')
                raise Exception(msg)
            """,
            'expected_errors': [],
        },
        {
            'code': """
                LOG = log.getLogger(__name__)
                msg = _('hello %s') % 'world'
                LOG.warning(msg)
            """,
            'expected_errors': [(3, 12, 'K005')],
        },
        {
            'code': """
                # this should not be an error
                LOG = log.getLogger(__name__)
                try:
                    something = True
                except AssertionError as e:
                    LOG.warning(e)
                    raise exception.Unauthorized(e)
            """,
            'expected_errors': [],
        },
        {
            'code': """
                # this should not be an error
                LOG = log.getLogger(__name__)
                try:
                    pass
                except AssertionError as e:
                    msg = _('some message')
                    LOG.warning(msg)
                    raise exception.Unauthorized(message=msg)
            """,
            'expected_errors': [(7, 16, 'K005')],
        },
    ]



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import http.client
from unittest import mock
import uuid

import fixtures
import freezegun
from oslo_db import exception as oslo_db_exception
from oslo_log import log
from oslo_utils import timeutils
from testtools import matchers

from keystone.common import provider_api
from keystone.common import sql
import keystone.conf
from keystone.credential.providers import fernet as credential_fernet
from keystone import exception
from keystone.identity.backends import base as identity_base
from keystone.identity.backends import resource_options as options
from keystone.identity.backends import sql_model as model
from keystone.tests import unit
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import database
from keystone.tests.unit import mapping_fixtures
from keystone.tests.unit import test_v3

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class IdentityTestCase(test_v3.RestfulTestCase):
    """Test users and groups."""

    def setUp(self):
        super().setUp()
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                credential_fernet.MAX_ACTIVE_KEYS,
            )
        )

        self.group = unit.new_group_ref(domain_id=self.domain_id)
        self.group = PROVIDERS.identity_api.create_group(self.group)
        self.group_id = self.group['id']

        self.credential = unit.new_credential_ref(
            user_id=self.user['id'], project_id=self.project_id
        )

        PROVIDERS.credential_api.create_credential(
            self.credential['id'], self.credential
        )

    # user crud tests

    def test_create_user(self):
        """Call ``POST /users``."""
        ref = unit.new_user_ref(domain_id=self.domain_id)
        r = self.post('/users', body={'user': ref})
        return self.assertValidUserResponse(r, ref)

    def test_create_user_without_domain(self):
        """Call ``POST /users`` without specifying domain.

        According to the identity-api specification, if you do not
        explicitly specific the domain_id in the entity, it should
        take the domain scope of the token as the domain_id.

        """
        # Create a user with a role on the domain so we can get a
        # domain scoped token
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        user = unit.create_user(PROVIDERS.identity_api, domain_id=domain['id'])
        PROVIDERS.assignment_api.create_grant(
            role_id=self.role_id, user_id=user['id'], domain_id=domain['id']
        )

        ref = unit.new_user_ref(domain_id=domain['id'])
        ref_nd = ref.copy()
        ref_nd.pop('domain_id')
        auth = self.build_authentication_request(
            user_id=user['id'],
            password=user['password'],
            domain_id=domain['id'],
        )
        r = self.post('/users', body={'user': ref_nd}, auth=auth)
        self.assertValidUserResponse(r, ref)

        # Now try the same thing without a domain token - which should fail
        ref = unit.new_user_ref(domain_id=domain['id'])
        ref_nd = ref.copy()
        ref_nd.pop('domain_id')
        auth = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )

        # TODO(henry-nash): Due to bug #1283539 we currently automatically
        # use the default domain_id if a domain scoped token is not being
        # used. For now we just check that a deprecation warning has been
        # issued. Change the code below to expect a failure once this bug is
        # fixed.
        with mock.patch(
            'oslo_log.versionutils.report_deprecated_feature'
        ) as mock_dep:
            r = self.post('/users', body={'user': ref_nd}, auth=auth)
            self.assertTrue(mock_dep.called)

        ref['domain_id'] = CONF.identity.default_domain_id
        return self.assertValidUserResponse(r, ref)

    def test_create_user_with_admin_token_and_domain(self):
        """Call ``POST /users`` with admin token and domain id."""
        ref = unit.new_user_ref(domain_id=self.domain_id)
        self.post(
            '/users',
            body={'user': ref},
            token=self.get_admin_token(),
            expected_status=http.client.CREATED,
        )

    def test_user_management_normalized_keys(self):
        """Illustrate the inconsistent handling of hyphens in keys.

        To quote Morgan in bug 1526244:

            the reason this is converted from "domain-id" to "domain_id" is
            because of how we process/normalize data. The way we have to handle
            specific data types for known columns requires avoiding "-" in the
            actual python code since "-" is not valid for attributes in python
            w/o significant use of "getattr" etc.

            In short, historically we handle some things in conversions. The
            use of "extras" has long been a poor design choice that leads to
            odd/strange inconsistent behaviors because of other choices made in
            handling data from within the body. (In many cases we convert from
            "-" to "_" throughout openstack)

        Source: https://bugs.launchpad.net/keystone/+bug/1526244/comments/9

        """
        # Create two domains to work with.
        domain1 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain1['id'], domain1)
        domain2 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain2['id'], domain2)

        # We can successfully create a normal user without any surprises.
        user = unit.new_user_ref(domain_id=domain1['id'])
        r = self.post('/users', body={'user': user})
        self.assertValidUserResponse(r, user)
        user['id'] = r.json['user']['id']

        # Query strings are not normalized: so we get all users back (like
        # self.user), not just the ones in the specified domain.
        r = self.get('/users?domain-id={}'.format(domain1['id']))
        self.assertValidUserListResponse(r, ref=self.user)
        self.assertNotEqual(domain1['id'], self.user['domain_id'])

        # When creating a new user, if we move the 'domain_id' into the
        # 'domain-id' attribute, the server will normalize the request
        # attribute, and effectively "move it back" for us.
        user = unit.new_user_ref(domain_id=domain1['id'])
        user['domain-id'] = user.pop('domain_id')
        r = self.post('/users', body={'user': user})
        self.assertNotIn('domain-id', r.json['user'])
        self.assertEqual(domain1['id'], r.json['user']['domain_id'])
        # (move this attribute back so we can use assertValidUserResponse)
        user['domain_id'] = user.pop('domain-id')
        self.assertValidUserResponse(r, user)
        user['id'] = r.json['user']['id']

        # If we try updating the user's 'domain_id' by specifying a
        # 'domain-id', then it'll be stored into extras rather than normalized,
        # and the user's actual 'domain_id' is not affected.
        r = self.patch(
            '/users/{}'.format(user['id']),
            body={'user': {'domain-id': domain2['id']}},
        )
        self.assertEqual(domain2['id'], r.json['user']['domain-id'])
        self.assertEqual(user['domain_id'], r.json['user']['domain_id'])
        self.assertNotEqual(domain2['id'], user['domain_id'])
        self.assertValidUserResponse(r, user)

    def test_create_user_bad_request(self):
        """Call ``POST /users``."""
        self.post(
            '/users',
            body={'user': {}},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_create_user_bad_domain_id(self):
        """Call ``POST /users``."""
        # create user with 'DEFaUlT' domain_id instead if 'default'
        # and verify it fails
        self.post(
            '/users',
            body={'user': {"name": "baddomain", "domain_id": "DEFaUlT"}},
            expected_status=http.client.NOT_FOUND,
        )

    def test_list_head_users(self):
        """Call ``GET & HEAD /users``."""
        resource_url = '/users'
        r = self.get(resource_url)
        self.assertValidUserListResponse(
            r, ref=self.user, resource_url=resource_url
        )
        self.head(resource_url, expected_status=http.client.OK)

    def test_list_users_with_multiple_backends(self):
        """Call ``GET /users`` when multiple backends is enabled.

        In this scenario, the controller requires a domain to be specified
        either as a filter or by using a domain scoped token.

        """
        self.config_fixture.config(
            group='identity', domain_specific_drivers_enabled=True
        )

        # Create a new domain with a new project and user
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)

        project = unit.new_project_ref(domain_id=domain['id'])
        PROVIDERS.resource_api.create_project(project['id'], project)

        user = unit.create_user(PROVIDERS.identity_api, domain_id=domain['id'])

        # Create both project and domain role grants for the user so we
        # can get both project and domain scoped tokens
        PROVIDERS.assignment_api.create_grant(
            role_id=self.role_id, user_id=user['id'], domain_id=domain['id']
        )
        PROVIDERS.assignment_api.create_grant(
            role_id=self.role_id, user_id=user['id'], project_id=project['id']
        )

        dom_auth = self.build_authentication_request(
            user_id=user['id'],
            password=user['password'],
            domain_id=domain['id'],
        )
        project_auth = self.build_authentication_request(
            user_id=user['id'],
            password=user['password'],
            project_id=project['id'],
        )

        # First try using a domain scoped token
        resource_url = '/users'
        r = self.get(resource_url, auth=dom_auth)
        self.assertValidUserListResponse(
            r, ref=user, resource_url=resource_url
        )

        # Now try using a project scoped token
        resource_url = '/users'
        r = self.get(resource_url, auth=project_auth)
        self.assertValidUserListResponse(
            r, ref=user, resource_url=resource_url
        )

        # Now try with an explicit filter
        resource_url = '/users?domain_id={domain_id}'.format(
            domain_id=domain['id']
        )
        r = self.get(resource_url)
        self.assertValidUserListResponse(
            r, ref=user, resource_url=resource_url
        )

    def test_list_users_no_default_project(self):
        """Call ``GET /users`` making sure no default_project_id."""
        user = unit.new_user_ref(self.domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        resource_url = '/users'
        r = self.get(resource_url)
        self.assertValidUserListResponse(
            r, ref=user, resource_url=resource_url
        )

    def test_get_head_user(self):
        """Call ``GET & HEAD /users/{user_id}``."""
        resource_url = '/users/{user_id}'.format(user_id=self.user['id'])
        r = self.get(resource_url)
        self.assertValidUserResponse(r, self.user)
        self.head(resource_url, expected_status=http.client.OK)

    def test_get_user_does_not_include_extra_attributes(self):
        """Call ``GET /users/{user_id}`` extra attributes are not included."""
        user = unit.new_user_ref(
            domain_id=self.domain_id, project_id=self.project_id
        )
        user = PROVIDERS.identity_api.create_user(user)
        self.assertNotIn('created_at', user)
        self.assertNotIn('last_active_at', user)

    def test_get_user_includes_required_attributes(self):
        """Call ``GET /users/{user_id}`` required attributes are included."""
        user = unit.new_user_ref(
            domain_id=self.domain_id, project_id=self.project_id
        )
        user = PROVIDERS.identity_api.create_user(user)
        self.assertIn('id', user)
        self.assertIn('name', user)
        self.assertIn('enabled', user)
        self.assertIn('password_expires_at', user)
        r = self.get('/users/{user_id}'.format(user_id=user['id']))
        self.assertValidUserResponse(r, user)

    def test_get_user_with_default_project(self):
        """Call ``GET /users/{user_id}`` making sure of default_project_id."""
        user = unit.new_user_ref(
            domain_id=self.domain_id, project_id=self.project_id
        )
        user = PROVIDERS.identity_api.create_user(user)
        r = self.get('/users/{user_id}'.format(user_id=user['id']))
        self.assertValidUserResponse(r, user)

    def test_add_user_to_group(self):
        """Call ``PUT /groups/{group_id}/users/{user_id}``."""
        self.put(
            '/groups/{group_id}/users/{user_id}'.format(
                group_id=self.group_id, user_id=self.user['id']
            )
        )

    def test_list_head_groups_for_user(self):
        """Call ``GET & HEAD /users/{user_id}/groups``."""
        user1 = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )
        user2 = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )

        self.put(
            '/groups/{group_id}/users/{user_id}'.format(
                group_id=self.group_id, user_id=user1['id']
            )
        )

        # Scenarios below are written to test the default policy configuration

        # One should be allowed to list one's own groups
        auth = self.build_authentication_request(
            user_id=user1['id'], password=user1['password']
        )
        resource_url = '/users/{user_id}/groups'.format(user_id=user1['id'])
        r = self.get(resource_url, auth=auth)
        self.assertValidGroupListResponse(
            r, ref=self.group, resource_url=resource_url
        )
        self.head(resource_url, auth=auth, expected_status=http.client.OK)

        # Administrator is allowed to list others' groups
        resource_url = '/users/{user_id}/groups'.format(user_id=user1['id'])
        r = self.get(resource_url)
        self.assertValidGroupListResponse(
            r, ref=self.group, resource_url=resource_url
        )
        self.head(resource_url, expected_status=http.client.OK)

        # Ordinary users should not be allowed to list other's groups
        auth = self.build_authentication_request(
            user_id=user2['id'], password=user2['password']
        )
        resource_url = '/users/{user_id}/groups'.format(user_id=user1['id'])
        self.get(
            resource_url,
            auth=auth,
            expected_status=exception.ForbiddenAction.code,
        )
        self.head(
            resource_url,
            auth=auth,
            expected_status=exception.ForbiddenAction.code,
        )

    def test_check_user_in_group(self):
        """Call ``HEAD /groups/{group_id}/users/{user_id}``."""
        self.put(
            '/groups/{group_id}/users/{user_id}'.format(
                group_id=self.group_id, user_id=self.user['id']
            )
        )
        self.head(
            '/groups/{group_id}/users/{user_id}'.format(
                group_id=self.group_id, user_id=self.user['id']
            )
        )

    def test_list_head_users_in_group(self):
        """Call ``GET & HEAD /groups/{group_id}/users``."""
        self.put(
            '/groups/{group_id}/users/{user_id}'.format(
                group_id=self.group_id, user_id=self.user['id']
            )
        )
        resource_url = f'/groups/{self.group_id}/users'
        r = self.get(resource_url)
        self.assertValidUserListResponse(
            r, ref=self.user, resource_url=resource_url
        )
        self.assertIn(
            f'/groups/{self.group_id}/users', r.result['links']['self']
        )
        self.head(resource_url, expected_status=http.client.OK)

    def test_remove_user_from_group(self):
        """Call ``DELETE /groups/{group_id}/users/{user_id}``."""
        self.put(
            '/groups/{group_id}/users/{user_id}'.format(
                group_id=self.group_id, user_id=self.user['id']
            )
        )
        self.delete(
            '/groups/{group_id}/users/{user_id}'.format(
                group_id=self.group_id, user_id=self.user['id']
            )
        )

    def test_update_ephemeral_user(self):
        federated_user_a = model.FederatedUser()
        federated_user_b = model.FederatedUser()
        federated_user_a.idp_id = 'a_idp'
        federated_user_b.idp_id = 'b_idp'
        federated_user_a.display_name = 'federated_a'
        federated_user_b.display_name = 'federated_b'
        federated_users = [federated_user_a, federated_user_b]

        user_a = model.User()
        user_a.federated_users = federated_users

        self.assertEqual(federated_user_a.display_name, user_a.name)
        self.assertIsNone(user_a.password)

        user_a.name = 'new_federated_a'

        self.assertEqual('new_federated_a', user_a.name)
        self.assertIsNone(user_a.local_user)

    def test_update_user(self):
        """Call ``PATCH /users/{user_id}``."""
        user = unit.new_user_ref(domain_id=self.domain_id)
        del user['id']
        r = self.patch(
            '/users/{user_id}'.format(user_id=self.user['id']),
            body={'user': user},
        )
        self.assertValidUserResponse(r, user)

    def test_admin_password_reset(self):
        # bootstrap a user as admin
        user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )

        # auth as user should work before a password change
        old_password_auth = self.build_authentication_request(
            user_id=user_ref['id'], password=user_ref['password']
        )
        r = self.v3_create_token(old_password_auth)
        old_token = r.headers.get('X-Subject-Token')

        # auth as user with a token should work before a password change
        old_token_auth = self.build_authentication_request(token=old_token)
        self.v3_create_token(old_token_auth)

        # administrative password reset
        new_password = uuid.uuid4().hex
        self.patch(
            '/users/{}'.format(user_ref['id']),
            body={'user': {'password': new_password}},
        )

        # auth as user with original password should not work after change
        self.v3_create_token(
            old_password_auth, expected_status=http.client.UNAUTHORIZED
        )

        # auth as user with an old token should not work after change
        self.v3_create_token(
            old_token_auth, expected_status=http.client.NOT_FOUND
        )

        # new password should work
        new_password_auth = self.build_authentication_request(
            user_id=user_ref['id'], password=new_password
        )
        self.v3_create_token(new_password_auth)

    def test_admin_password_reset_with_min_password_age_enabled(self):
        # enable minimum_password_age, this should have no effect on admin
        # password reset
        self.config_fixture.config(
            group='security_compliance', minimum_password_age=1
        )
        # create user
        user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )
        # administrative password reset
        new_password = uuid.uuid4().hex
        r = self.patch(
            '/users/{}'.format(user_ref['id']),
            body={'user': {'password': new_password}},
        )
        self.assertValidUserResponse(r, user_ref)
        # authenticate with new password
        new_password_auth = self.build_authentication_request(
            user_id=user_ref['id'], password=new_password
        )
        self.v3_create_token(new_password_auth)

    def test_admin_password_reset_with_password_lock(self):
        # create user
        user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )
        lock_pw_opt = options.LOCK_PASSWORD_OPT.option_name
        update_user_body = {'user': {'options': {lock_pw_opt: True}}}
        self.patch('/users/{}'.format(user_ref['id']), body=update_user_body)

        # administrative password reset
        new_password = uuid.uuid4().hex
        r = self.patch(
            '/users/{}'.format(user_ref['id']),
            body={'user': {'password': new_password}},
        )
        self.assertValidUserResponse(r, user_ref)
        # authenticate with new password
        new_password_auth = self.build_authentication_request(
            user_id=user_ref['id'], password=new_password
        )
        self.v3_create_token(new_password_auth)

    def test_update_user_domain_id(self):
        """Call ``PATCH /users/{user_id}`` with domain_id.

        A user's `domain_id` is immutable. Ensure that any attempts to update
        the `domain_id` of a user fails.
        """
        user = unit.new_user_ref(domain_id=self.domain['id'])
        user = PROVIDERS.identity_api.create_user(user)
        user['domain_id'] = CONF.identity.default_domain_id
        self.patch(
            '/users/{user_id}'.format(user_id=user['id']),
            body={'user': user},
            expected_status=exception.ValidationError.code,
        )

    def test_delete_user(self):
        """Call ``DELETE /users/{user_id}``.

        As well as making sure the delete succeeds, we ensure
        that any credentials that reference this user are
        also deleted, while other credentials are unaffected.
        In addition, no tokens should remain valid for this user.

        """
        # First check the credential for this user is present
        r = PROVIDERS.credential_api.get_credential(self.credential['id'])
        self.assertDictEqual(self.credential, r)
        # Create a second credential with a different user

        user2 = unit.new_user_ref(
            domain_id=self.domain['id'], project_id=self.project['id']
        )
        user2 = PROVIDERS.identity_api.create_user(user2)
        credential2 = unit.new_credential_ref(
            user_id=user2['id'], project_id=self.project['id']
        )
        PROVIDERS.credential_api.create_credential(
            credential2['id'], credential2
        )

        # Create a token for this user which we can check later
        # gets deleted
        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )
        token = self.get_requested_token(auth_data)
        # Confirm token is valid for now
        self.head(
            '/auth/tokens',
            headers={'X-Subject-Token': token},
            expected_status=http.client.OK,
        )

        # Now delete the user
        self.delete('/users/{user_id}'.format(user_id=self.user['id']))

        # Deleting the user should have deleted any credentials
        # that reference this project
        self.assertRaises(
            exception.CredentialNotFound,
            PROVIDERS.credential_api.get_credential,
            self.credential['id'],
        )
        # But the credential for user2 is unaffected
        r = PROVIDERS.credential_api.get_credential(credential2['id'])
        self.assertDictEqual(credential2, r)

    def test_delete_user_retries_on_deadlock(self):
        patcher = mock.patch(
            'sqlalchemy.orm.query.Query.delete', autospec=True
        )

        class FakeDeadlock:
            def __init__(self, mock_patcher):
                self.deadlock_count = 2
                self.mock_patcher = mock_patcher
                self.patched = True

            def __call__(self, *args, **kwargs):
                if self.deadlock_count > 1:
                    self.deadlock_count -= 1
                else:
                    self.mock_patcher.stop()
                    self.patched = False
                raise oslo_db_exception.DBDeadlock

        sql_delete_mock = patcher.start()
        side_effect = FakeDeadlock(patcher)
        sql_delete_mock.side_effect = side_effect

        user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )

        try:
            PROVIDERS.identity_api.delete_user(user_id=user_ref['id'])
        finally:
            if side_effect.patched:
                patcher.stop()

        call_count = sql_delete_mock.call_count

        # initial attempt + 1 retry
        delete_user_attempt_count = 2
        self.assertEqual(call_count, delete_user_attempt_count)

    # group crud tests

    def test_create_group(self):
        """Call ``POST /groups``."""
        # Create a new group to avoid a duplicate check failure
        ref = unit.new_group_ref(domain_id=self.domain_id)
        r = self.post('/groups', body={'group': ref})
        return self.assertValidGroupResponse(r, ref)

    def test_create_group_bad_request(self):
        """Call ``POST /groups``."""
        self.post(
            '/groups',
            body={'group': {}},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_list_head_groups(self):
        """Call ``GET & HEAD /groups``."""
        resource_url = '/groups'
        r = self.get(resource_url)
        self.assertValidGroupListResponse(
            r, ref=self.group, resource_url=resource_url
        )
        self.head(resource_url, expected_status=http.client.OK)

    def test_get_head_group(self):
        """Call ``GET & HEAD /groups/{group_id}``."""
        resource_url = f'/groups/{self.group_id}'
        r = self.get(resource_url)
        self.assertValidGroupResponse(r, self.group)
        self.head(resource_url, expected_status=http.client.OK)

    def test_update_group(self):
        """Call ``PATCH /groups/{group_id}``."""
        group = unit.new_group_ref(domain_id=self.domain_id)
        del group['id']
        r = self.patch(f'/groups/{self.group_id}', body={'group': group})
        self.assertValidGroupResponse(r, group)

    def test_update_group_domain_id(self):
        """Call ``PATCH /groups/{group_id}`` with domain_id.

        A group's `domain_id` is immutable. Ensure that any attempts to update
        the `domain_id` of a group fails.
        """
        self.group['domain_id'] = CONF.identity.default_domain_id
        self.patch(
            '/groups/{group_id}'.format(group_id=self.group['id']),
            body={'group': self.group},
            expected_status=exception.ValidationError.code,
        )

    def test_delete_group(self):
        """Call ``DELETE /groups/{group_id}``."""
        self.delete(f'/groups/{self.group_id}')

    def test_create_user_password_not_logged(self):
        # When a user is created, the password isn't logged at any level.

        log_fix = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))

        ref = unit.new_user_ref(domain_id=self.domain_id)
        self.post('/users', body={'user': ref})

        self.assertNotIn(ref['password'], log_fix.output)

    def test_update_password_not_logged(self):
        # When admin modifies user password, the password isn't logged at any
        # level.

        log_fix = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))

        # bootstrap a user as admin
        user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )

        self.assertNotIn(user_ref['password'], log_fix.output)

        # administrative password reset
        new_password = uuid.uuid4().hex
        self.patch(
            '/users/{}'.format(user_ref['id']),
            body={'user': {'password': new_password}},
        )

        self.assertNotIn(new_password, log_fix.output)

    def test_setting_default_project_id_to_domain_failed(self):
        """Call ``POST and PATCH /users`` default_project_id=domain_id.

        Make sure we validate the default_project_id if it is specified.
        It cannot be set to a domain_id, even for a project acting as domain
        right now. That's because we haven't sort out the issuing
        project-scoped token for project acting as domain bit yet. Once we
        got that sorted out, we can relax this constraint.

        """
        # creating a new user with default_project_id set to a
        # domain_id should result in HTTP 400
        ref = unit.new_user_ref(
            domain_id=self.domain_id, project_id=self.domain_id
        )
        self.post(
            '/users',
            body={'user': ref},
            token=CONF.admin_token,
            expected_status=http.client.BAD_REQUEST,
        )

        # updating user's default_project_id to a domain_id should result
        # in HTTP 400
        user = {'default_project_id': self.domain_id}
        self.patch(
            '/users/{user_id}'.format(user_id=self.user['id']),
            body={'user': user},
            token=CONF.admin_token,
            expected_status=http.client.BAD_REQUEST,
        )


class ChangePasswordTestCase(test_v3.RestfulTestCase):
    def setUp(self):
        super().setUp()
        self.user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )
        self.token = self.get_request_token(
            self.user_ref['password'], http.client.CREATED
        )

    def get_request_token(self, password, expected_status):
        auth_data = self.build_authentication_request(
            user_id=self.user_ref['id'], password=password
        )
        r = self.v3_create_token(auth_data, expected_status=expected_status)
        return r.headers.get('X-Subject-Token')

    def change_password(self, expected_status, **kwargs):
        """Return a test response for a change password request."""
        return self.post(
            '/users/{}/password'.format(self.user_ref['id']),
            body={'user': kwargs},
            token=self.token,
            expected_status=expected_status,
        )


class UserSelfServiceChangingPasswordsTestCase(ChangePasswordTestCase):
    def _create_user_with_expired_password(self):
        expire_days = CONF.security_compliance.password_expires_days + 1
        time = timeutils.utcnow() - datetime.timedelta(expire_days)
        password = uuid.uuid4().hex
        user_ref = unit.new_user_ref(
            domain_id=self.domain_id, password=password
        )
        with freezegun.freeze_time(time):
            self.user_ref = PROVIDERS.identity_api.create_user(user_ref)

        return password

    def test_changing_password(self):
        # original password works
        token_id = self.get_request_token(
            self.user_ref['password'], expected_status=http.client.CREATED
        )
        # original token works
        old_token_auth = self.build_authentication_request(token=token_id)
        self.v3_create_token(old_token_auth)

        # change password
        new_password = uuid.uuid4().hex
        self.change_password(
            password=new_password,
            original_password=self.user_ref['password'],
            expected_status=http.client.NO_CONTENT,
        )

        # old password fails
        self.get_request_token(
            self.user_ref['password'], expected_status=http.client.UNAUTHORIZED
        )

        # old token fails
        self.v3_create_token(
            old_token_auth, expected_status=http.client.NOT_FOUND
        )

        # new password works
        self.get_request_token(
            new_password, expected_status=http.client.CREATED
        )

    def test_changing_password_with_min_password_age(self):
        time = timeutils.utcnow()
        with freezegun.freeze_time(time) as frozen_datetime:
            # enable minimum_password_age and attempt to change password
            new_password = uuid.uuid4().hex
            self.config_fixture.config(
                group='security_compliance', minimum_password_age=1
            )
            # able to change password after create user
            self.change_password(
                password=new_password,
                original_password=self.user_ref['password'],
                expected_status=http.client.NO_CONTENT,
            )
            # 2nd change password should fail due to minimum password age and
            # make sure we wait one second to avoid race conditions with Fernet
            frozen_datetime.tick(delta=datetime.timedelta(seconds=1))
            self.token = self.get_request_token(
                new_password, http.client.CREATED
            )
            self.change_password(
                password=uuid.uuid4().hex,
                original_password=new_password,
                expected_status=http.client.BAD_REQUEST,
            )
            # disable minimum_password_age and attempt to change password
            self.config_fixture.config(
                group='security_compliance', minimum_password_age=0
            )
            self.change_password(
                password=uuid.uuid4().hex,
                original_password=new_password,
                expected_status=http.client.NO_CONTENT,
            )

    def test_changing_password_with_password_lock(self):
        password = uuid.uuid4().hex
        ref = unit.new_user_ref(domain_id=self.domain_id, password=password)
        response = self.post('/users', body={'user': ref})
        user_id = response.json_body['user']['id']

        time = timeutils.utcnow()
        with freezegun.freeze_time(time) as frozen_datetime:
            # Lock the user's password
            lock_pw_opt = options.LOCK_PASSWORD_OPT.option_name
            user_patch = {'user': {'options': {lock_pw_opt: True}}}
            self.patch(f'/users/{user_id}', body=user_patch)

            # Fail, password is locked
            new_password = uuid.uuid4().hex
            body = {
                'user': {
                    'original_password': password,
                    'password': new_password,
                }
            }
            path = f'/users/{user_id}/password'
            self.post(path, body=body, expected_status=http.client.BAD_REQUEST)

            # Unlock the password, and change should work
            user_patch['user']['options'][lock_pw_opt] = False
            self.patch(f'/users/{user_id}', body=user_patch)

            path = f'/users/{user_id}/password'
            self.post(path, body=body, expected_status=http.client.NO_CONTENT)

            frozen_datetime.tick(delta=datetime.timedelta(seconds=1))

            auth_data = self.build_authentication_request(
                user_id=user_id, password=new_password
            )
            self.v3_create_token(
                auth_data, expected_status=http.client.CREATED
            )

            path = f'/users/{user_id}'
            user = self.get(path).json_body['user']
            self.assertIn(lock_pw_opt, user['options'])
            self.assertFalse(user['options'][lock_pw_opt])

            # Completely unset the option from the user's reference
            user_patch['user']['options'][lock_pw_opt] = None
            self.patch(f'/users/{user_id}', body=user_patch)
            path = f'/users/{user_id}'
            user = self.get(path).json_body['user']
            self.assertNotIn(lock_pw_opt, user['options'])

    def test_changing_password_with_missing_original_password_fails(self):
        r = self.change_password(
            password=uuid.uuid4().hex, expected_status=http.client.BAD_REQUEST
        )
        self.assertThat(
            r.result['error']['message'],
            matchers.Contains('original_password'),
        )

    def test_changing_password_with_missing_password_fails(self):
        r = self.change_password(
            original_password=self.user_ref['password'],
            expected_status=http.client.BAD_REQUEST,
        )
        self.assertThat(
            r.result['error']['message'], matchers.Contains('password')
        )

    def test_changing_password_with_incorrect_password_fails(self):
        self.change_password(
            password=uuid.uuid4().hex,
            original_password=uuid.uuid4().hex,
            expected_status=http.client.UNAUTHORIZED,
        )

    def test_changing_password_with_disabled_user_fails(self):
        # disable the user account
        self.user_ref['enabled'] = False
        self.patch(
            '/users/{}'.format(self.user_ref['id']),
            body={'user': self.user_ref},
        )

        self.change_password(
            password=uuid.uuid4().hex,
            original_password=self.user_ref['password'],
            expected_status=http.client.UNAUTHORIZED,
        )

    def test_changing_password_not_logged(self):
        # When a user changes their password, the password isn't logged at any
        # level.

        log_fix = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))

        # change password
        new_password = uuid.uuid4().hex
        self.change_password(
            password=new_password,
            original_password=self.user_ref['password'],
            expected_status=http.client.NO_CONTENT,
        )

        self.assertNotIn(self.user_ref['password'], log_fix.output)
        self.assertNotIn(new_password, log_fix.output)

    def test_changing_expired_password_succeeds(self):
        self.config_fixture.config(
            group='security_compliance', password_expires_days=2
        )
        password = self._create_user_with_expired_password()

        new_password = uuid.uuid4().hex
        self.change_password(
            password=new_password,
            original_password=password,
            expected_status=http.client.NO_CONTENT,
        )
        # new password works
        self.get_request_token(
            new_password, expected_status=http.client.CREATED
        )

    def test_changing_expired_password_with_disabled_user_fails(self):
        self.config_fixture.config(
            group='security_compliance', password_expires_days=2
        )

        password = self._create_user_with_expired_password()
        # disable the user account
        self.user_ref['enabled'] = False
        self.patch(
            '/users/{}'.format(self.user_ref['id']),
            body={'user': self.user_ref},
        )

        new_password = uuid.uuid4().hex
        self.change_password(
            password=new_password,
            original_password=password,
            expected_status=http.client.UNAUTHORIZED,
        )

    def test_change_password_required_upon_first_use_for_create(self):
        self.config_fixture.config(
            group='security_compliance', change_password_upon_first_use=True
        )

        # create user
        self.user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )

        # attempt to authenticate with create user password
        self.get_request_token(
            self.user_ref['password'], expected_status=http.client.UNAUTHORIZED
        )

        # self-service change password
        new_password = uuid.uuid4().hex
        self.change_password(
            password=new_password,
            original_password=self.user_ref['password'],
            expected_status=http.client.NO_CONTENT,
        )

        # authenticate with the new password
        self.token = self.get_request_token(new_password, http.client.CREATED)

    def test_change_password_required_upon_first_use_for_admin_reset(self):
        self.config_fixture.config(
            group='security_compliance', change_password_upon_first_use=True
        )

        # admin reset
        reset_password = uuid.uuid4().hex
        user_password = {'password': reset_password}
        PROVIDERS.identity_api.update_user(self.user_ref['id'], user_password)

        # attempt to authenticate with admin reset password
        self.get_request_token(
            reset_password, expected_status=http.client.UNAUTHORIZED
        )

        # self-service change password
        new_password = uuid.uuid4().hex
        self.change_password(
            password=new_password,
            original_password=reset_password,
            expected_status=http.client.NO_CONTENT,
        )

        # authenticate with the new password
        self.token = self.get_request_token(new_password, http.client.CREATED)

    def test_change_password_required_upon_first_use_ignore_user(self):
        self.config_fixture.config(
            group='security_compliance', change_password_upon_first_use=True
        )

        # ignore user and reset password
        reset_password = uuid.uuid4().hex
        self.user_ref['password'] = reset_password
        ignore_opt_name = options.IGNORE_CHANGE_PASSWORD_OPT.option_name
        self.user_ref['options'][ignore_opt_name] = True
        PROVIDERS.identity_api.update_user(self.user_ref['id'], self.user_ref)

        # authenticate with the reset password
        self.token = self.get_request_token(
            reset_password, http.client.CREATED
        )

    def test_lockout_exempt(self):
        self.config_fixture.config(
            group='security_compliance', lockout_failure_attempts=1
        )

        # create user
        self.user_ref = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )

        # update the user, mark her as exempt from lockout
        ignore_opt_name = options.IGNORE_LOCKOUT_ATTEMPT_OPT.option_name
        self.user_ref['options'][ignore_opt_name] = True
        PROVIDERS.identity_api.update_user(self.user_ref['id'], self.user_ref)

        # fail to auth, this should lockout the user, since we're allowed
        # one failure, but we're exempt from lockout!
        bad_password = uuid.uuid4().hex
        self.token = self.get_request_token(
            bad_password, http.client.UNAUTHORIZED
        )

        # attempt to authenticate with correct password
        self.get_request_token(
            self.user_ref['password'], expected_status=http.client.CREATED
        )


class PasswordValidationTestCase(ChangePasswordTestCase):
    def setUp(self):
        super().setUp()
        # passwords requires: 1 letter, 1 digit, 7 chars
        self.config_fixture.config(
            group='security_compliance',
            password_regex=(r'^(?=.*\d)(?=.*[a-zA-Z]).{7,}$'),
        )

    def test_create_user_with_invalid_password(self):
        user = unit.new_user_ref(domain_id=self.domain_id)
        user['password'] = 'simple'
        self.post(
            '/users',
            body={'user': user},
            token=self.get_admin_token(),
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_user_with_invalid_password(self):
        user = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain['id']
        )
        user['password'] = 'simple'
        self.patch(
            '/users/{user_id}'.format(user_id=user['id']),
            body={'user': user},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_changing_password_with_simple_password_strength(self):
        # password requires: any non-whitespace character
        self.config_fixture.config(
            group='security_compliance', password_regex=r'[\S]+'
        )
        self.change_password(
            password='simple',
            original_password=self.user_ref['password'],
            expected_status=http.client.NO_CONTENT,
        )

    def test_changing_password_with_strong_password_strength(self):
        self.change_password(
            password='mypassword2',
            original_password=self.user_ref['password'],
            expected_status=http.client.NO_CONTENT,
        )

    def test_changing_password_with_strong_password_strength_fails(self):
        # no digit
        self.change_password(
            password='mypassword',
            original_password=self.user_ref['password'],
            expected_status=http.client.BAD_REQUEST,
        )

        # no letter
        self.change_password(
            password='12345678',
            original_password=self.user_ref['password'],
            expected_status=http.client.BAD_REQUEST,
        )

        # less than 7 chars
        self.change_password(
            password='mypas2',
            original_password=self.user_ref['password'],
            expected_status=http.client.BAD_REQUEST,
        )


class UserFederatedAttributesTests(test_v3.RestfulTestCase):
    def _create_federated_attributes(self):
        # Create the idp
        idp = {
            'id': uuid.uuid4().hex,
            'enabled': True,
            'description': uuid.uuid4().hex,
        }
        PROVIDERS.federation_api.create_idp(idp['id'], idp)
        # Create the mapping
        mapping = mapping_fixtures.MAPPING_EPHEMERAL_USER
        mapping['id'] = uuid.uuid4().hex
        PROVIDERS.federation_api.create_mapping(mapping['id'], mapping)
        # Create the protocol
        protocol = {'id': uuid.uuid4().hex, 'mapping_id': mapping['id']}
        PROVIDERS.federation_api.create_protocol(
            idp['id'], protocol['id'], protocol
        )
        return idp, protocol

    def _create_user_with_federated_user(self, user, fed_dict):
        with sql.session_for_write() as session:
            federated_ref = model.FederatedUser.from_dict(fed_dict)
            user_ref = model.User.from_dict(user)
            user_ref.created_at = timeutils.utcnow()
            user_ref.federated_users.append(federated_ref)
            session.add(user_ref)
            return identity_base.filter_user(user_ref.to_dict())

    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.load_backends()
        # Create the federated object
        idp, protocol = self._create_federated_attributes()
        self.fed_dict = unit.new_federated_user_ref()
        self.fed_dict['idp_id'] = idp['id']
        self.fed_dict['protocol_id'] = protocol['id']
        self.fed_dict['unique_id'] = "jdoe"
        # Create the domain_id, user, and federated_user relationship
        self.domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(self.domain['id'], self.domain)
        self.fed_user = unit.new_user_ref(domain_id=self.domain['id'])
        self.fed_user = self._create_user_with_federated_user(
            self.fed_user, self.fed_dict
        )
        # Create two new fed_users which will have the same idp and protocol
        # but be completely different from the first fed_user
        # Create a new idp and protocol for fed_user2 and 3
        idp, protocol = self._create_federated_attributes()
        self.fed_dict2 = unit.new_federated_user_ref()
        self.fed_dict2['idp_id'] = idp['id']
        self.fed_dict2['protocol_id'] = protocol['id']
        self.fed_dict2['unique_id'] = "ravelar"
        self.fed_user2 = unit.new_user_ref(domain_id=self.domain['id'])
        self.fed_user2 = self._create_user_with_federated_user(
            self.fed_user2, self.fed_dict2
        )
        self.fed_dict3 = unit.new_federated_user_ref()
        self.fed_dict3['idp_id'] = idp['id']
        self.fed_dict3['protocol_id'] = protocol['id']
        self.fed_dict3['unique_id'] = "jsmith"
        self.fed_user3 = unit.new_user_ref(domain_id=self.domain['id'])
        self.fed_user3 = self._create_user_with_federated_user(
            self.fed_user3, self.fed_dict3
        )

    def _test_list_users_with_federated_parameter(self, parameter):
        # construct the resource url based off what's passed in parameter
        resource_url = f'/users?{parameter[0]}={self.fed_dict[parameter[0]]}'
        for attr in parameter[1:]:
            resource_url += f'&{attr}={self.fed_dict[attr]}'
        r = self.get(resource_url)
        # Check that only one out of 3 fed_users is matched by calling the api
        # and that it is a valid response
        self.assertEqual(1, len(r.result['users']))
        self.assertValidUserListResponse(
            r, ref=self.fed_user, resource_url=resource_url
        )
        # Since unique_id will always return one user if matching for unique_id
        # in the query, we rule out unique_id for the next tests
        if not any('unique_id' in x for x in parameter):
            # Check that we get two matches here since fed_user2 and fed_user3
            # both have the same idp and protocol
            resource_url = (
                f'/users?{parameter[0]}={self.fed_dict2[parameter[0]]}'
            )
            for attr in parameter[1:]:
                resource_url += f'&{attr}={self.fed_dict2[attr]}'
            r = self.get(resource_url)
            self.assertEqual(2, len(r.result['users']))
            self.assertValidUserListResponse(
                r, ref=self.fed_user2, resource_url=resource_url
            )

    def test_list_users_with_idp_id(self):
        attribute = ['idp_id']
        self._test_list_users_with_federated_parameter(attribute)

    def test_list_users_with_protocol_id(self):
        attribute = ['protocol_id']
        self._test_list_users_with_federated_parameter(attribute)

    def test_list_users_with_unique_id(self):
        attribute = ['unique_id']
        self._test_list_users_with_federated_parameter(attribute)

    def test_list_users_with_idp_id_and_unique_id(self):
        attribute = ['idp_id', 'unique_id']
        self._test_list_users_with_federated_parameter(attribute)

    def test_list_users_with_idp_id_and_protocol_id(self):
        attribute = ['idp_id', 'protocol_id']
        self._test_list_users_with_federated_parameter(attribute)

    def test_list_users_with_protocol_id_and_unique_id(self):
        attribute = ['protocol_id', 'unique_id']
        self._test_list_users_with_federated_parameter(attribute)

    def test_list_users_with_all_federated_attributes(self):
        attribute = ['idp_id', 'protocol_id', 'unique_id']
        self._test_list_users_with_federated_parameter(attribute)

    def test_get_user_includes_required_federated_attributes(self):
        user = self.identity_api.get_user(self.fed_user['id'])
        self.assertIn('federated', user)
        self.assertIn('idp_id', user['federated'][0])
        self.assertIn('protocols', user['federated'][0])
        self.assertIn('protocol_id', user['federated'][0]['protocols'][0])
        self.assertIn('unique_id', user['federated'][0]['protocols'][0])
        r = self.get('/users/{user_id}'.format(user_id=user['id']))
        self.assertValidUserResponse(r, user)

    def test_create_user_with_federated_attributes(self):
        """Call ``POST /users``."""
        idp, protocol = self._create_federated_attributes()
        ref = unit.new_user_ref(domain_id=self.domain_id)
        ref['federated'] = [
            {
                'idp_id': idp['id'],
                'protocols': [
                    {
                        'protocol_id': protocol['id'],
                        'unique_id': uuid.uuid4().hex,
                    }
                ],
            }
        ]
        r = self.post('/users', body={'user': ref})
        user = r.result['user']
        self.assertEqual(user['name'], ref['name'])
        self.assertEqual(user['federated'], ref['federated'])
        self.assertValidUserResponse(r, ref)

    def test_create_user_fails_when_given_invalid_idp_and_protocols(self):
        """Call ``POST /users`` with invalid idp and protocol to fail."""
        idp, protocol = self._create_federated_attributes()
        ref = unit.new_user_ref(domain_id=self.domain_id)
        ref['federated'] = [
            {
                'idp_id': 'fakeidp',
                'protocols': [
                    {
                        'protocol_id': 'fakeprotocol_id',
                        'unique_id': uuid.uuid4().hex,
                    }
                ],
            }
        ]

        self.post(
            '/users',
            body={'user': ref},
            token=self.get_admin_token(),
            expected_status=http.client.BAD_REQUEST,
        )
        ref['federated'][0]['idp_id'] = idp['id']
        self.post(
            '/users',
            body={'user': ref},
            token=self.get_admin_token(),
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_user_with_federated_attributes(self):
        """Call ``PATCH /users/{user_id}``."""
        user = self.fed_user.copy()
        del user['id']
        user['name'] = 'James Doe'
        idp, protocol = self._create_federated_attributes()
        user['federated'] = [
            {
                'idp_id': idp['id'],
                'protocols': [
                    {'protocol_id': protocol['id'], 'unique_id': 'jdoe'}
                ],
            }
        ]
        r = self.patch(
            '/users/{user_id}'.format(user_id=self.fed_user['id']),
            body={'user': user},
        )
        resp_user = r.result['user']
        self.assertEqual(user['name'], resp_user['name'])
        self.assertEqual(user['federated'], resp_user['federated'])
        self.assertValidUserResponse(r, user)

    def test_update_user_fails_when_given_invalid_idp_and_protocols(self):
        """Call ``PATCH /users/{user_id}``."""
        user = self.fed_user.copy()
        del user['id']
        idp, protocol = self._create_federated_attributes()
        user['federated'] = [
            {
                'idp_id': 'fakeidp',
                'protocols': [
                    {
                        'protocol_id': 'fakeprotocol_id',
                        'unique_id': uuid.uuid4().hex,
                    }
                ],
            }
        ]

        self.patch(
            '/users/{user_id}'.format(user_id=self.fed_user['id']),
            body={'user': user},
            expected_status=http.client.BAD_REQUEST,
        )
        user['federated'][0]['idp_id'] = idp['id']
        self.patch(
            '/users/{user_id}'.format(user_id=self.fed_user['id']),
            body={'user': user},
            expected_status=http.client.BAD_REQUEST,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.auth import plugins
from keystone.tests import unit


class TestPluginCore(unit.TestCase):
    def test_construct_method_map_with_one_methods(self):
        auth_methods = ['password']
        self.config_fixture.config(group='auth', methods=auth_methods)

        expected_method_map = {1: 'password'}
        method_map = plugins.construct_method_map_from_config()
        self.assertDictEqual(expected_method_map, method_map)

    def test_construct_method_map_with_two_methods(self):
        auth_methods = ['password', 'token']
        self.config_fixture.config(group='auth', methods=auth_methods)

        expected_method_map = {1: 'password', 2: 'token'}
        method_map = plugins.construct_method_map_from_config()
        self.assertDictEqual(expected_method_map, method_map)

    def test_construct_method_map_with_three_methods(self):
        auth_methods = ['password', 'token', 'totp']
        self.config_fixture.config(group='auth', methods=auth_methods)

        expected_method_map = {1: 'password', 2: 'token', 4: 'totp'}
        method_map = plugins.construct_method_map_from_config()
        self.assertDictEqual(expected_method_map, method_map)

    def test_convert_methods_to_integer(self):
        auth_methods = ['password', 'token', 'totp']
        self.config_fixture.config(group='auth', methods=auth_methods)

        method_integer = plugins.convert_method_list_to_integer(['password'])
        self.assertEqual(1, method_integer)

        method_integer = plugins.convert_method_list_to_integer(
            ['password', 'token']
        )
        self.assertEqual(3, method_integer)

        method_integer = plugins.convert_method_list_to_integer(
            ['password', 'totp']
        )
        self.assertEqual(5, method_integer)

        method_integer = plugins.convert_method_list_to_integer(
            ['token', 'totp']
        )
        self.assertEqual(6, method_integer)

        method_integer = plugins.convert_method_list_to_integer(
            ['password', 'token', 'totp']
        )
        self.assertEqual(7, method_integer)

    def test_convert_integer_to_methods(self):
        auth_methods = ['password', 'token', 'totp']
        self.config_fixture.config(group='auth', methods=auth_methods)

        expected_methods = ['password']
        methods = plugins.convert_integer_to_method_list(1)
        self.assertTrue(len(methods) == 1)
        for method in methods:
            self.assertIn(method, expected_methods)

        expected_methods = ['password', 'token']
        methods = plugins.convert_integer_to_method_list(3)
        self.assertTrue(len(methods) == 2)
        for method in methods:
            self.assertIn(method, expected_methods)

        expected_methods = ['password', 'totp']
        methods = plugins.convert_integer_to_method_list(5)
        self.assertTrue(len(methods) == 2)
        for method in methods:
            self.assertIn(method, expected_methods)

        expected_methods = ['token', 'totp']
        methods = plugins.convert_integer_to_method_list(6)
        self.assertTrue(len(methods) == 2)
        for method in methods:
            self.assertIn(method, expected_methods)

        expected_methods = ['password', 'token', 'totp']
        methods = plugins.convert_integer_to_method_list(7)
        self.assertTrue(len(methods) == 3)
        for method in methods:
            self.assertIn(method, expected_methods)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Before request processing for JSON Body enforcement

import flask
from werkzeug import exceptions as werkzeug_exceptions

from keystone import exception
from keystone.i18n import _
from keystone.server.flask import common as ks_flask_common


def json_body_before_request():
    """Enforce JSON Request Body."""
    # TODO(morgan): Allow other content-types when OpenAPI Doc or improved
    # federation is implemented for known/valid paths. This function should
    # be removed long term.

    # exit if there is nothing to be done, (no body)
    if not flask.request.get_data():
        return None
    elif flask.request.path and flask.request.path.startswith(
        '/v3/OS-OAUTH2/'
    ):
        # When the user makes a request to the OAuth2.0 token endpoint,
        # the user should use the "application/x-www-form-urlencoded" format
        # with a character encoding of UTF-8 in the HTTP request entity-body.
        # At the scenario there is nothing to be done and exit.
        return None

    try:
        # flask does loading for us for json, use the flask default loader
        # in the case that the data is *not* json or a dict, we should see a
        # raise of werkzeug.exceptions.BadRequest, re-spin this to the keystone
        # ValidationError message (as expected by our contract)

        # Explicitly check if the content is supposed to be json.
        if (
            flask.request.is_json
            or flask.request.headers.get('Content-Type', '') == ''
        ):
            json_decoded = flask.request.get_json(force=True)
            if not isinstance(json_decoded, dict):
                # In the case that the returned value was not a dict, force
                # a raise that will be caught the same way that a Decode error
                # would be handled.
                raise werkzeug_exceptions.BadRequest(
                    _('resulting JSON load was not a dict')
                )
        else:
            # We no longer need enforcement on this API, set unenforced_ok
            # we already hit a validation error. This is required as the
            # request is never hitting the resource methods, meaning
            # @unenforced_api is not called. Without marking the request
            # as "unenforced_ok" the assertion check to ensure enforcement
            # was called would raise up causing a 500 error.
            ks_flask_common.set_unenforced_ok()
            raise exception.ValidationError(
                attribute='application/json', target='Content-Type header'
            )

    except werkzeug_exceptions.BadRequest:
        # We no longer need enforcement on this API, set unenforced_ok
        # we already hit a validation error. This is required as the
        # request is never hitting the resource methods, meaning
        # @unenforced_api is not called. Without marking the request
        # as "unenforced_ok" the assertion check to ensure enforcement
        # was called would raise up causing a 500 error.
        ks_flask_common.set_unenforced_ok()
        raise exception.ValidationError(
            attribute='valid JSON', target='request body'
        )
